/********************************************************/
/*                                                      */
/*               OO7 Benchmark                          */
/*                                                      */
/*              COPYRIGHT (C) 1993                      */
/*                                                      */
/*                Michael J. Carey 		        */
/*                David J. DeWitt 		        */
/*                Jeffrey Naughton 		        */
/*               Madison, WI U.S.A.                     */
/*                                                      */
/*	         ALL RIGHTS RESERVED                    */
/*                                                      */
/********************************************************/


////////////////////////////////////////////////////////////////////////////
//
// CompositePart Methods
//
////////////////////////////////////////////////////////////////////////////

#include <libc.h>
#include <string.h>

#include <ostore/ostore.hh>
#include <ostore/coll.hh>
#include <ostore/coll/iplist.hh>
#include <ostore/relat.hh>
#include "OO7.h"
#include "GenParams.h"
#include "VarParams.h"
#include "baidlist.h"

extern os_database* oo7db;

extern int	nextAtomicId;
extern int	debugMode;
extern char*	types[NumTypes];

extern BAIdList* private_cp;
extern BAIdList* shared_cp;


////////////////////////////////////////////////////////////////////////////
//
// CompositePart Constructor
//
////////////////////////////////////////////////////////////////////////////

CompositePart::CompositePart(int cpId) : id(cpId)
{
   BaseAssembly* ba;

   if (debugMode) {
        printf("BEGIN CompositePart::CompositePart(cpId = %d)\n", cpId);
   }

    // workaround for ODI bag bug
    usedInPriv.change_behavior(usedInPriv.get_behavior() |
   			       os_collection::allow_duplicates);
    usedInShar.change_behavior(usedInShar.get_behavior() |
    			       os_collection::allow_duplicates);

   if (debugMode) {
        printf("IN CompositePart::CompositePart:  init data members\n");
   }

    // initialize this composite part's simple data members (some randomly)

    int typeNo = (int) (random() % NumTypes);
    strncpy(type, types[typeNo], TypeSize);

    // for the build date, decide if this part is young or old, and then
    // randomly choose a date in the required range

    if (cpId % YoungCompFrac == 0) {
        // young one
        if (debugMode) {
            fprintf(stderr, "(young composite part, id = %d.)\n", id);
        }
        buildDate = MinYoungCompDate + 
		    (int)(random() % (MaxYoungCompDate - MinYoungCompDate + 1));
    } else { 
        // old one
        if (debugMode) {
            fprintf(stderr, "(old composite part, id = %d.)\n", id);
        }
        buildDate = MinOldCompDate + 
		    (int)(random() % (MaxOldCompDate - MinOldCompDate + 1));
    }

   if (debugMode) {
        printf("IN CompositePart::CompositePart:  update extent\n");
   }

    // add part to CompositePart extent

    (*CompositePart::extent) |= this;

   if (debugMode) {
        printf("IN CompositePart::CompositePart:  create document\n");
   }

    // initialize the documentation (indexed by its title) ...

    documentation = new (os_segment::of(Document::extent)) Document (cpId);

   if (debugMode) {
        printf("IN CompositePart::CompositePart:  create atomic parts\n");
   }

    // now create the atomic parts (indexed by their ids) ...

    int i, id, from, to;
    AtomicPart** atom = new AtomicPart* [NumAtomicPerComp];
    Connection* conn;

    for (i = 0; i < NumAtomicPerComp; i++) {
	id = nextAtomicId + i;
	atom[i] = new (os_segment::of(this), this) AtomicPart (id);
	parts |= atom[i];
	if (i == 0) { rootPart = atom[i]; }
    }

   if (debugMode) {
        printf("IN CompositePart::CompositePart:  wire up via connections\n");
   }

    // ... and then wire them semi-randomly together (as a ring plus random
    // additional connections to ensure full part reachability for traversals)

    for (from = 0; from < NumAtomicPerComp; from++) {
	for (i = 0; i < NumConnPerAtomic; i++) {
	    if (i == 0) {
	        to = (from + 1) % NumAtomicPerComp;
	    } else {
	        to = (int) (random() % NumAtomicPerComp);
	    }
	    conn = new (os_segment::of(this), this)
	    	   Connection (atom[from], atom[to]);
	}
    }

   if (debugMode) {
        printf("IN CompositePart::CompositePart:  local cleanup\n");
   }

    delete [NumAtomicPerComp] atom;
    nextAtomicId += NumAtomicPerComp;

    // finally insert this composite part as a child of the base
    // assemblies that use it

    // first the assemblies using the comp part as a shared component
    shared_cp[cpId].initScan();

    // get the first base assembly
    ba = shared_cp[cpId].next();
    while (ba != NULL)
    {
        // add this composite part cp the list of shared parts used
        // in the assembly
	ba->componentsShar |= this;

        // continue with next base assembly
        ba = shared_cp[cpId].next();
    }

    // next the assemblies using the comp part as a private component
    private_cp[cpId].initScan();

    // get the first base assembly
    ba = private_cp[cpId].next();
    while (ba != NULL)
    {
	// add this composite part to the list of private parts used
        // in the assembly
	ba->componentsPriv |= this;

        // continue with next base assembly
        ba = private_cp[cpId].next();
    }

   if (debugMode) {
        printf("END CompositePart::CompositePart(cpId = %d)\n", cpId);
   }
}


////////////////////////////////////////////////////////////////////////////
//
// CompositePart Destructor
//
////////////////////////////////////////////////////////////////////////////

CompositePart::~CompositePart()
{
    if (debugMode) {
        printf("Entering CompositePart::~CompositePart(id = %d)\n", id);
    }

    // delete associated atomic parts
if (debugMode) {
    printf("    In ~CompositePart, about to delete parts\n");
}
    AtomicPart* p;
    while (!parts.empty()) {
	p = parts.pick();
        delete p;
    }

    // delete associated documentation
if (debugMode) {
    printf("    In ~CompositePart, about to delete documentation\n");
}
    delete (Document*) documentation;

    // employ workaround for ODI index maintenance bug
    // __back_ptr.make_link(&documentation, &documentation,
    // 			    -os_index(CompositePart,documentation));

    // remove part from CompositePart extent
if (debugMode) {
    printf("    In ~CompositePart, about to remove part from CompositePart::extent\n");
}
    (*CompositePart::extent) -= this;

    if (debugMode) {
        printf("Leaving CompositePart::~CompositePart(id = %d)\n", id);
    }
}

