/********************************************************/
/*                                                      */
/*               OO7 Benchmark                          */
/*                                                      */
/*              COPYRIGHT (C) 1993                      */
/*                                                      */
/*                Michael J. Carey 		        */
/*                David J. DeWitt 		        */
/*                Jeffrey Naughton 		        */
/*               Madison, WI U.S.A.                     */
/*                                                      */
/*	         ALL RIGHTS RESERVED                    */
/*                                                      */
/********************************************************/

#include <libc.h>
#include <string.h>
#include <stdio.h>
#include <sys/resource.h>
#include <stdlib.h> 

#include <ostore/ostore.hh>
#include <ostore/coll.hh>
#include <ostore/coll/iplist.hh>
#include <ostore/relat.hh>
#include "OO7.h"
#include "GenParams.h"
#include "VarParams.h"
#include "baidlist.h"


//////////////////////////////////////////////////////////////////
//
// Global Variables, etc., for Benchmark Database Generation
//
//////////////////////////////////////////////////////////////////

struct rusage startUsage, endUsage;
struct timeval startWallTime;
struct timeval endWallTime;
struct timezone ignoreTimeZone;


extern void SetParams(char* configFileName);

extern double ComputeWallClockTime(struct timeval *startWallTime, 
	           struct timeval *endWallTime);

extern double ComputeUserTime(struct rusage *startUsage, 
	           struct rusage *endUsage);

extern double ComputeSystemTime(struct rusage *startUsage, 
	           struct rusage *endUsage);

os_database* oo7db;

int	nextAtomicId = 1;
int	nextCompositeId = 1;
int	nextComplexAssemblyId = 1;
int	nextBaseAssemblyId = 1;
int	nextModuleId = 1;
int	debugMode = FALSE;

char    *types[NumTypes] = {
		"type000", "type001", "type002", "type003", "type004",
		"type005", "type006", "type007", "type008", "type009"
	};

// initializers for various type extent pointers

persistent<oo7db> os_Set<AtomicPart*>* AtomicPart::extent = 0;

persistent<oo7db> os_Set<Document*>* Document::extent = 0;

persistent<oo7db> os_Set<CompositePart*>* CompositePart::extent = 0; 

persistent<oo7db> os_Set<BaseAssembly*>* BaseAssembly::extent = 0;

persistent<oo7db> os_Set<Module*>* Module::extent = 0;

// data structures to keep track of which composite parts each base
// assemblies uses.  This is done so that we can generate the base
// assemblies first and then the composite parts.  Why bother?
// In particular, if you did the composite parts first and the base
// assemblies, there would be no need for these data structures.
// This second approach worked fine for the small and medium databases
// but not for the large as the program seeked all over the disk.

BAIdList* private_cp;
BAIdList* shared_cp;

// usage:   GenDB <configFileName> [-d]

main(int argc, char** argv)
{
	if (argc < 2) 
	{
	      fprintf(stderr, "Usage: %s <configFileName> [-d]\n", argv[0]);
	      exit(1);
	}
	// See if debug mode is desired
	if ((argc == 3) && (strcmp(argv[2], "-d") == 0)) debugMode = TRUE;
	else  debugMode = FALSE;
	
	SetParams(argv[1]);

	shared_cp = new BAIdList[TotalCompParts+1];
	private_cp = new BAIdList[TotalCompParts+1];

	objectstore::set_cache_size(CachePages * 8192);

	printf("Initialized system with call objectstore::set_cache_size(%d)\n",
	        CachePages * 8192);

        // get wall clock time
        gettimeofday(&startWallTime, &ignoreTimeZone);

	
	//////////////////////////////////////////////////////////////////
	//
	// Initialize ObjectStore
	//
	//////////////////////////////////////////////////////////////////


	// Start the transaction

	transaction* xact = os_transaction::begin();

	// get starting usage values.
	getrusage(RUSAGE_SELF, &startUsage);


	//////////////////////////////////////////////////////////////////
	//
	// Replace and Open the Database
	//
	//////////////////////////////////////////////////////////////////

	char *dbname = getenv("DBNAME");
	if (! dbname) {
		dbname = "/n/sparc6/raw/data_fs/os_db";
	}
	oo7db = os_database::create(dbname, 0664, 1);
	if (oo7db == NULL) {
	      fprintf(stderr, "ERROR: Cannot open OO7 Database.\n");
	      os_transaction::abort();
	      exit(1);
	}
	oo7db->set_read_whole_segment(0);

	//////////////////////////////////////////////////////////////////
	//
	// Do Benchmark Data Generation
	//
	//////////////////////////////////////////////////////////////////

        // Atomic and Composite parts
	os_segment* part_seg = oo7db->create_segment();
        int part_seg_size = 
	  ((TotalConnections * sizeof(Connection))
	   + (TotalAtomicParts * (sizeof(AtomicPart)))
	   + (TotalCompParts * (sizeof(CompositePart))));
	part_seg_size = (int) (1.2 * (float) part_seg_size);
	part_seg->set_size(part_seg_size);
	AtomicPart::extent = &os_Set<AtomicPart*>::create(part_seg);
	CompositePart::extent = &os_Set<CompositePart*>::create(part_seg); 
        
        // Documents
	os_segment* doc_seg;
        doc_seg = oo7db->create_segment();
	int doc_seg_size = TotalModules * ManualSize + 
		(TotalCompParts * (sizeof(Document) + DocumentSize));
	doc_seg_size = (int) (1.2 * (float) doc_seg_size);
	doc_seg->set_size(doc_seg_size);
	Document::extent = &os_Set<Document*>::create(doc_seg);

	// Assemblies and Modules 
	BaseAssembly::extent = 
                       &os_Set<BaseAssembly*>::create(oo7db->default_segment);
	Module::extent = &os_Set<Module*>::create(oo7db->default_segment);

        // set random number seed.
        srandom(1);


	// First generate the desired number of modules, each in its own
	// segment
	while (nextModuleId <= TotalModules) 
	{
	    printf("Generating module %d.\n", nextModuleId);
	    int id = nextModuleId++;
	    Module* module = new (oo7db->default_segment) Module (id);
	}

	// Generate the desired number of composite parts in a design
	// library segment, with each part being in its own cluster

        int id;
	int clusterSize;
        CompositePart* comp;

	clusterSize = int (1.1 * (NumAtomicPerComp * (sizeof(AtomicPart) +
			          (NumConnPerAtomic * sizeof(Connection)))));

	while (nextCompositeId <= TotalCompParts) 
	{
	    id = nextCompositeId++;

            if (debugMode) {
                 printf(">>> creating composite part %d itself\n", id);
            }

	    comp = new (part_seg) CompositePart (id);

            if (debugMode) {
                 printf(">>> done creating composite part %d\n", id);
            }

	    if ((nextCompositeId % 500) == 0) 
	    {
/*
		printf("commit xact\n");
		os_transaction::commit(xact);
*/
		printf("created %d composite parts \n",nextCompositeId);
/*
		xact = os_transaction::begin();
		printf("start a new transaction\n");
*/
	    }
        }

	// Index the extent of CompositeParts 
	os_segment* part_index_seg;
        part_index_seg = oo7db->create_segment();
	CompositePart::extent->add_index(pathof(CompositePart*, id),
            part_index_seg
        );

	printf("created composite parts index\n");

        // Create indexes on ...
        // ... Parts
	AtomicPart::extent->add_index(pathof(AtomicPart*, id), part_index_seg);
	AtomicPart::extent->add_index(pathof(AtomicPart*, buildDate),
				      os_collection::ordered, 
				      part_index_seg);

	printf("created atomic parts index\n");

        // .... and Documents
	os_segment* doc_index_seg;
        doc_index_seg = oo7db->create_segment();
	Document::extent->add_index(pathof(Document*, title), doc_index_seg);
	Document::extent->add_index(pathof(Document*, id), doc_index_seg);

	printf("created document index\n");

        // .... and finally, Assemblies and Modules
	BaseAssembly::extent->add_index(
	//      pathof(BaseAssembly*, componentsPriv[]->documentation->title)
	    os_index_path::create(
	        "BaseAssembly*",
	        "componentsPriv[]->documentation->title",
	        os_database::get_transient_database()
	    )
	);
	BaseAssembly::extent->add_index(pathof(BaseAssembly*, id));
	Module::extent->add_index(pathof(Module*, id));

	printf("created assembly index\n");

	// Print out some useful information about what happened

	printf("=== DONE CREATING DATABASE, TOTALS WERE ===\n");
	printf("    # atomic parts\t\t%d\n", nextAtomicId-1);
	printf("    # composite parts\t\t%d\n", nextCompositeId-1);
	printf("    # complex assemblies\t%d\n", nextComplexAssemblyId-1);
	printf("    # base assemblies\t\t%d\n", nextBaseAssemblyId-1);
	printf("    # modules\t\t\t%d\n", nextModuleId-1);

	//////////////////////////////////////////////////////////////////
	//
	// Shutdown ObjectStore
	//
	//////////////////////////////////////////////////////////////////


	// Commit the transaction
	os_transaction::commit(xact);

        // compute and report wall clock time
        gettimeofday(&endWallTime, &ignoreTimeZone);
        printf("Wall-Clock time to generate database: %f seconds.\n", 
	             ComputeWallClockTime(&startWallTime, &endWallTime));


	getrusage(RUSAGE_SELF, &endUsage);
	fprintf(stdout, "(CPU: %f seconds.)\n", 
	                ComputeUserTime(&startUsage, &endUsage) +
			ComputeSystemTime(&startUsage, &endUsage));


//      // create 60,000 1000 byte dummy objects
//	char* purgeVar;
//	purgeVar = getenv("PURGE");
//	if (purgeVar != NULL) {
//	    if (atoi(purgeVar) == 1) {
//        	printf("creating 60000 dummy 1,000 byte objects\n");
//		CreateDummy (60000);
//	    }
//	    else printf("not creating dummy purge collection\n");
//	}
//	else printf("not creating dummy purge collection\n");

	// Close the Database
	oo7db->close();

	// Exit
	exit(0);
}

