/********************************************************/
/*                                                      */
/*               OO7 Benchmark                          */
/*                                                      */
/*              COPYRIGHT (C) 1993                      */
/*                                                      */
/*                Michael J. Carey 		        */
/*                David J. DeWitt 		        */
/*                Jeffrey Naughton 		        */
/*               Madison, WI U.S.A.                     */
/*                                                      */
/*	         ALL RIGHTS RESERVED                    */
/*                                                      */
/********************************************************/

#include <libc.h>
#include <string.h>
#include <stdio.h>
#include <unistd.h>

#include <ostore/ostore.hh>
#include <ostore/coll.hh>
#include <ostore/relat.hh>
#include "OO7.h"
#include "BenchParams.h"
#include "GenParams.h"
#include "VarParams.h"

//////////////////////////////////////////////////////////////////
//
// Traverse - DFS traverse module.  Upon reaching a base assembly,
// visit all referenced composite parts.  At each composite part,
// take an action that depends on which traversal variant has been
// requested.
//
///////////////////////////////////////////////////////////////////

extern os_database* oo7db;

extern int	debugMode;

extern void 	PrintOp(BenchmarkOp op);

//////////////////////////////////////////////////////////////////
//
// Module Method for Traversal
//
//////////////////////////////////////////////////////////////////


int Module::traverse(BenchmarkOp op)
{
    if (debugMode) {
        printf("Module::traverse(id = %d, op = ", id);
        PrintOp(op);
        printf(")\n");
    }

    // now traverse the assembly hierarchy
    return designRoot->traverse(op);
}


//////////////////////////////////////////////////////////////////
//
// ComplexAssembly Method for Traversal
//
//////////////////////////////////////////////////////////////////

int ComplexAssembly::traverse(BenchmarkOp op)
{
    if (debugMode) {
        printf("\tComplexAssembly::traverse(id = %d, op = ", id);
        PrintOp(op);
        printf(")\n");
    }

    // prepare for and execute the traversal
    int count = 0;
    Assembly* assm;

    if (op < MultiTrav1) {

        // fully traverse the assembly hierarchy
        foreach (assm, subAssemblies) {
            count += assm->traverse(op);
        }

    } else {

        // randomly walk down the assembly hierarchy
	int curPath = 0;
	int randPath = (int) (random() % NumAssmPerAssm);
        foreach (assm, subAssemblies) {
	    if (curPath++ == randPath) {
                count += assm->traverse(op);
		break;
	    }
        }
    }

    // lastly, return the result
    return count;
}


//////////////////////////////////////////////////////////////////
//
// BaseAssembly Method for Traversal
//
//////////////////////////////////////////////////////////////////


int BaseAssembly::traverse(BenchmarkOp op)
{
    if (debugMode) {
        printf("\t\tBaseAssembly::traverse(id = %d, op = ", id);
        PrintOp(op);
        printf(")\n");
    }

    // prepare for and execute the traversal
    int count = 0;
    CompositePart* comp;

    if (op < MultiTrav1) {

        // fully traverse each of the assembly's private components
        foreach (comp, componentsPriv) {
            count += comp->traverse(op);
        }

    } else {

        // first, traverse private and/or shared composite parts

	if (op < MultiTrav3 || op > MultiTrav4) {
            foreach (comp, componentsPriv) {
                count += comp->traverse(Trav1);
            }
	}

	if (op > MultiTrav2) {
            foreach (comp, componentsShar) {
                count += comp->traverse(Trav1);
            }
	}

        // next, sleep for awhile to reduce server disk contention

	sleep(MultiSleepTime);

        // also, perform an update traversal if one is desired

	if (op == MultiTrav2 || op == MultiTrav5 || op == MultiTrav6) {
            foreach (comp, componentsPriv) {
                count += comp->traverse(Trav2b);
            }
	} else if (op == MultiTrav4) {
            foreach (comp, componentsShar) {
                count += comp->traverse(Trav2b);
            }
	}

    }

    // lastly, return the result

    return count;
}


//////////////////////////////////////////////////////////////////
//
// CompositePart Method for Traversal
//
//////////////////////////////////////////////////////////////////


int CompositePart::traverse(BenchmarkOp op)
{
    if (debugMode) {
        printf("\t\t\tCompositePart::traverse(id = %d, op = ", id);
        PrintOp(op);
        printf(")\n");
    }

    if ((op >= Trav1) && (op <= Trav3c)) {

        // do parameterized DFS of atomic part graph
        PartIdSet visitedIds;
        return rootPart->traverse(op, visitedIds);

    } else if (op == Trav4) {

        // search document text for a certain character
        return documentation->searchText('I');

    } else if (op == Trav5do) {

        // conditionally change initial part of document text
        return documentation->replaceText("I am", "This is");

    } else if (op == Trav5undo) {

        // conditionally change back initial part of document text
        return documentation->replaceText("This is", "I am");

    } else if (op == Trav6) {
	
	// visit the root part only (it knows how to handle this)
        PartIdSet visitedIds;
        return rootPart->traverse(op, visitedIds);

    } else {

	// composite parts don't respond to other traversals
	printf("*** CompositePart::PANIC -- illegal traversal!!! ***\n");

    }
}


//////////////////////////////////////////////////////////////////
//
// AtomicPart Method for Traversal
//
//////////////////////////////////////////////////////////////////


int AtomicPart::traverse(BenchmarkOp op, PartIdSet& visitedIds)
{
    if (debugMode) {
        printf("\t\t\t\tAtomicPart::traverse(id = %d, op = ", id);
        PrintOp(op);
        printf(")\n");
    }

    int i;
    int count = 0;

    switch (op) {

	case Trav1:

	    // just examine the part
            count += 1;
            DoNothing();
	    break;

        case Trav2a:

	    // swap X and Y if first part
	    if (visitedIds.empty()) { 
                swapXY(); 
                count += 1;
            }
	    break;

        case Trav2b:

	    // swap X and Y
            swapXY();
            count += 1;
	    break;

        case Trav2c:

	    // swap X and Y repeatedly
            for (i = 0; i < UpdateRepeatCnt; i++) { 
               swapXY(); 
               count += 1;
            }
	    break;

        case Trav3a:

	    // toggle date if first part
	    if (visitedIds.empty()) { 
                toggleDate(); 
                count += 1;
            }
	    break;

        case Trav3b:

	    // toggle date
            toggleDate();
            count += 1;
	    break;

        case Trav3c:

	    // toggle date repeatedly
            for (i = 0; i < UpdateRepeatCnt; i++) { 
               toggleDate();
               count += 1;
            }
	    break;

	case Trav6:

	    // examine only the root part
            count += 1;
            DoNothing();
	    return count;

        default:

	    // atomic parts don't respond to other traversals
	    printf("*** AtomicPart::PANIC -- illegal traversal!!! ***\n");

    }

    // now, record the fact that we've visited this part
    visitedIds.insert(id);

    // finally, continue with DFS of atomic parts graph
    Connection* conn;
    foreach (conn, to) {
        if (!visitedIds.contains(conn->to->id)) {
	    count += conn->to->traverse(op, visitedIds);
        }
    }
    return count;
}

